pub mod raas {
    pub mod register {
        include!(concat!(env!("OUT_DIR"), "/raas.register.rs"));
    }

    pub mod bot {
        pub mod roll {
            include!(concat!(env!("OUT_DIR"), "/raas.bot.roll.rs"));
        }
    }

    pub mod ping {
        include!(concat!(env!("OUT_DIR"), "/raas.ping.rs"));
    }

    pub mod cmd {
        include!(concat!(env!("OUT_DIR"), "/raas.cmd.rs"));
    }

    pub mod resp {
        include!(concat!(env!("OUT_DIR"), "/raas.resp.rs"));
    }

    #[derive(Debug, Clone)]
    pub struct RaasMessage {
        pub len: u32,
        pub msg: Vec<u8>,
    }

    impl RaasMessage {
        pub fn new(msg: Vec<u8>) -> Self {
            Self {
                len: msg.len() as u32,
                msg,
            }
        }

        pub fn from_bytes(bytes: &[u8]) -> Self {
            let (len_bytes, msg) = bytes.split_at(std::mem::size_of::<u32>());
            let mut arr = [0u8; 4];
            arr.copy_from_slice(len_bytes);
            let len = u32::from_be_bytes(arr);

            Self {
                len,
                msg: msg.to_vec(),
            }
        }

        pub fn into_bytes(mut self) -> Vec<u8> {
            let mut bytes = Vec::new();

            bytes.extend_from_slice(&self.len.to_be_bytes());

            bytes.append(&mut self.msg);

            bytes
        }
    }
}

#[cfg(test)]
mod test {
    use crate::raas;
    use crate::raas::RaasMessage;

    #[test]
    fn create_cmd() {
        let cmd = raas::cmd::Command {
            id: 0,
            timestamp: 0,
            cmd: Some(raas::cmd::command::Cmd::RollCmd(raas::bot::roll::RollCmd {
                cmd: Some(raas::bot::roll::roll_cmd::Cmd::Roll(
                    raas::bot::roll::Roll { rotations: 1 },
                )),
            })),
        };

        assert_eq!(cmd.id, 0);
        assert_eq!(cmd.timestamp, 0);
        assert!(cmd.cmd.is_some());
    }

    #[test]
    fn encode_message() {
        let data = vec![0, 1, 2, 3, 4];
        let msg = RaasMessage::new(data.clone());

        let msg_bytes = msg.into_bytes();

        assert_eq!(msg_bytes.len(), data.len() + std::mem::size_of::<u32>());
    }

    #[test]
    fn decode_message() {
        let data: Vec<u8> = vec![0, 0, 0, 5, 0, 1, 2, 3, 4];
        let msg = RaasMessage::from_bytes(&data);

        assert_eq!(msg.len, 5);
        assert_eq!(msg.msg, data[4..]);
    }

    #[test]
    fn encode_decode_message() {
        let original_message = RaasMessage::new(vec![0x55, 0xff]);

        let data = original_message.clone().into_bytes();

        let decoded_message = RaasMessage::from_bytes(&data);

        assert_eq!(decoded_message.len, decoded_message.len);
        assert_eq!(decoded_message.msg, original_message.msg);
    }
}
